#ifndef _KBD_VALUES_H
#define _KBD_VALUES_H

/* $Id$ */

/* Macros to be used by kernel internals (keyboard.c, vt.c) 
 * and by applications (loadkeys) to
 * handle keysym values in keymaps (Unicode, 8bit pairs; store, exchange).
 * imz@altlinux.org, Apr 2004 */
/* Inspired by patches for Unicode in console collected by Bruno Haible */

#include <asm/types.h>

/* There are 2 parts in this file: Storage, Kernel-User Interface. */

/******************** Storage *******************/
/* These def's are used in the kernel to store values in keymaps.
 * The same can be used in user applications for the same goal.
 */

/* There must be *macros* for stroring a value :-(, 
 * (and not only a static inline func or whatever else)
 * because they are supposed to be used in initializers
 * (initail keymaps). This makes things uglier. 
 */

/* Defining the area to strore 8-bit pairs: TYP-VAL: */

/* The storage method relies on the fact that KTYP can be represented by 4 bits: */
/* #if ( NR_TYPES > 0xf ) */
/* #error too many types, do not fit in 4 bits; this breaks the storage method */
/* #endif */

#define K_AREA_8BIT_BASE 0xe000
#define K_AREA_8BIT_ENDEXT  0xf000

/* Characteristic function of the area: */
static __inline__ int /* bool */ K_AREA_8BIT(const __u16 keysym)
{
	return (keysym >= K_AREA_8BIT_BASE) && (keysym < K_AREA_8BIT_ENDEXT);
}

/* Should only be applied to keysym for which K_AREA_8BIT(keysym) is true,
 * otherwise it's nonsense. Returns a TYP-VAL pair. */
static __inline__ __u16 K_RETR_8BIT(const __u16 keysym)
{
	return keysym - K_AREA_8BIT_BASE;
}

#define K8(ktypval) ((ktypval) + K_AREA_8BIT_BASE)

static __inline__ __u16 K_STORE_8BIT(__u16 ktypval)
{
	return K8(ktypval);
}

/* Used to be like this: */
#ifndef __KERNEL__ /* only user side may need the data about the past */

# define K_AREA_8BIT_BASE_OLD 0xf000
# define K8_OLD(ktypval) ((ktypval) + K_AREA_8BIT_BASE_OLD)

static __inline__ __u16 K_STORE_8BIT_OLD(const __u16 ktypval)
{ 
  return K8_OLD(ktypval);
}

#endif

/* Definitions of "Letter with case; subject to CapsLock" markers
 * which can be set on Unicode values: */

/* Private area where we store the marked values  */
#define K_AREA_UNI_WITHCASE_BASE 0xf000
#define K_AREA_UNI_WITHCASE_ENDEXT 0xf900
static __inline__ int /* bool */ K_AREA_UNI_WITHCASE(const __u16 keysym)
{
	return (keysym >= K_AREA_UNI_WITHCASE_BASE)
		&& (keysym < K_AREA_UNI_WITHCASE_ENDEXT);
}

/* Maps from public areas which have case 
 * into offsetted areas inside the greater private area: */
  /* *_BASE, *_END, *_WITHCASE_OFFSET */
  /* 0xf000..0xf58f -> U+0000..U+058f (Latin, Greek, Cyrillic, Armenian), */
#define K_AREA_UNI_MAIN_SCRIPTS_BASE 0x0000
#define K_AREA_UNI_MAIN_SCRIPTS_END 0x058f
#define K_AREA_UNI_MAIN_SCRIPTS_WITHCASE_OFFSET 0x0000
  /* 0xf5a0..0xf5ff -> U+10a0..U+10ff (Georgian), */
#define K_AREA_UNI_GEORGIAN_BASE 0x10a0
#define K_AREA_UNI_GEORGIAN_END 0x10ff
#define K_AREA_UNI_GEORGIAN_WITHCASE_OFFSET 0x05a0
  /* 0xf600..0xf7ff -> U+1e00..U+1fff (extended Latin & Greek), */
#define K_AREA_UNI_EXTEND_LATGREEK_BASE 0x1e00
#define K_AREA_UNI_EXTEND_LATGREEK_END 0x1fff
#define K_AREA_UNI_EXTEND_LATGREEK_WITHCASE_OFFSET 0x0600
  /* 0xf800..0xf8ef -> U+ff00..U+ffef (fullwidth forms: Latin). */
#define K_AREA_UNI_FULLW_LATIN_BASE 0xff00
#define K_AREA_UNI_FULLW_LATIN_END 0xffef
#define K_AREA_UNI_FULLW_LATIN_WITHCASE_OFFSET 0x0800

/* There are two variants of defining the complete map
 * consisting of the individual pieces for each script:
 *
 * either by putting the individual map parametres into a single array,
 * and then iterating through the array when doing calculations,
 *
 * or by defining equivalent macros for all these things (which is not 
 * as nice as the array variant, but is neccessary because we use
 * this map in initial values which must be constants).
*/

/* define K_AREAS_WITHCASE_ARRAY 
 * if you want to include the array variant 
 */

#ifdef K_AREAS_WITHCASE_ARRAY

# define WITHCASE_MAP(name) \
  { K_AREA_UNI_ ## name ## _BASE, \
    K_AREA_UNI_ ## name ## _END, \
    K_AREA_UNI_ ## name ## _WITHCASE_OFFSET \
  }
static const struct { __u16 base, end, offset; } 
areas_withcase[] = {
	WITHCASE_MAP(MAIN_SCRIPTS),
	WITHCASE_MAP(GEORGIAN),
	WITHCASE_MAP(EXTEND_LATGREEK),
	WITHCASE_MAP(FULLW_LATIN)
};
# undef WITHCASE_MAP
# define NR_AREAS_WITHCASE \
  ( sizeof(areas_withcase) / sizeof(areas_withcase[0]) )

#endif /* end if K_AREAS_WITHCASE_ARRAY */

#define K_UNI_WITHCASE_TRY(name, c) \
  ( ((c) >= K_AREA_UNI_ ## name ## _BASE) \
    && ((c) <= K_AREA_UNI_ ## name ## _END) ) \
  ? (c) - K_AREA_UNI_ ## name ## _BASE \
    + K_AREA_UNI_WITHCASE_BASE \
    + K_AREA_UNI_ ## name ## _WITHCASE_OFFSET

#define K_UNILETTER(c) ( \
	K_UNI_WITHCASE_TRY(MAIN_SCRIPTS, c) \
	: K_UNI_WITHCASE_TRY(GEORGIAN, c) \
	: K_UNI_WITHCASE_TRY(EXTEND_LATGREEK, c) \
	: K_UNI_WITHCASE_TRY(FULLW_LATIN, c) \
        : (c) /* not possible to mark; pass through without any changes */ \
)

static __inline__ __u16 K_UNI_ADD_WITHCASE(const __u16 c) {
/* Two equivalent variants to define the function */
#ifndef K_AREAS_WITHCASE_ARRAY
	return K_UNILETTER(c);
#else
	int i;
	for (i = 0; i < NR_AREAS_WITHCASE; ++i)
		if ( (c >= areas_withcase[i].base) 
		     && (c <= areas_withcase[i].end) )
		{
			return c - areas_withcase[i].base
				+ K_AREA_UNI_WITHCASE_BASE 
				+ areas_withcase[i].offset;
		}
	return c; /* not possible to mark; pass through without any changes */
#endif
}

#define K_UNI_WITHCASE_TRY_INVERSE(name, c) \
  ( ((c) >= K_AREA_UNI_ ## name ## _WITHCASE_OFFSET) \
    && ((c) <= K_AREA_UNI_ ## name ## _WITHCASE_OFFSET \
	+ K_AREA_UNI_ ## name ## _END \
        - K_AREA_UNI_ ## name ## _BASE) ) \
  ? (c) - K_AREA_UNI_ ## name ## _WITHCASE_OFFSET \
        + K_AREA_UNI_ ## name ## _BASE 

#define K_UNI_WITHCASE_TRY_ALL_INVERSE(c) ( \
	K_UNI_WITHCASE_TRY_INVERSE(MAIN_SCRIPTS, c) \
	: K_UNI_WITHCASE_TRY_INVERSE(GEORGIAN, c) \
	: K_UNI_WITHCASE_TRY_INVERSE(EXTEND_LATGREEK, c) \
	: K_UNI_WITHCASE_TRY_INVERSE(FULLW_LATIN, c) \
        : K_AREA_UNI_WITHCASE_BASE + (c) /* not possible to mark; pass through without any changes */ \
)

#define K_UNILETTER_INVERSE(c) ( \
        (! K_AREA_UNI_WITHCASE(c) ) \
        ? (c) /* no mark */ \
	: K_UNI_WITHCASE_TRY_ALL_INVERSE(c - K_AREA_UNI_WITHCASE_BASE) \
)

static __inline__ __u16 K_UNI_STRIP_WITHCASE(__u16 c) {
/* Two equivalent variants to define the function */
#ifndef K_AREAS_WITHCASE_ARRAY
	return K_UNILETTER_INVERSE(c);
#else
	unsigned int i;
	if (! K_AREA_UNI_WITHCASE(c) )
		return c; /* no mark */
	c -= K_AREA_UNI_WITHCASE_BASE;
	for (i = 0; i < NR_AREAS_WITHCASE; ++i)
		if ( (c >= areas_withcase[i].offset) 
		     && (c <= areas_withcase[i].offset 
			 + areas_withcase[i].end - areas_withcase[i].base) )
		{
			return c - areas_withcase[i].offset + areas_withcase[i].base;
		}
	/* What to do if it is an illegal code? Well, pass it through. */
	return K_AREA_UNI_WITHCASE_BASE + c; /* this should not happen */
#endif
}

/* K_AREA_8BIT and K_AREA_UNI_WITHCASE cover the whole Unicode Private Use Area. */

/* Used not to have such a marker for any Unicode characters: */
#ifndef __KERNEL__ /* only user side may need the data about the past */

# define K_UNILETTER_OLD(c) (c)

static __inline__ __u16 K_UNI_ADD_WITHCASE_OLD(const __u16 c) {
	return K_UNILETTER_OLD(c);
}

#endif


/******************** Kernel-User Interface  ************************/

/* Internally, we store the Unicode chracters (to which keys are mapped)
 * as their Unicode number represented by 2 bytes. 
 * But they are not exchanged in a transparent way 
 * between the user and the kernel (the interface used in vt.c)
 * (AFAIU due to compatibility with older versions):
 * The special (TYP,VAL) pairs should be exchanged not as private
 * Unicode values (the internal representation), but "transparently",
 * i.e. as pairs of numbers, making up the range of values in 0x0000--0x0f00.
 * We define the conversion of general Unicode values 
 * used by the current interface 
 * by the pair of macros SUBMIT_UNI, ACCEPT_UNI (must be a one-to-one map).
 * The same conversion rules should be used on the kernel side (vt.c)
 * and on the user side (loadkeys(1)); these macros can be exported for that
 * purpose. 
 */

/* All the interfaces rely on the fact that KTYP can be represented by 4 bits: */
/* #if ( NR_TYPES > 0xf ) */
/* #error too many types, do not fit in 4 bits; this breaks the interface conversion */
/* #endif */

/* The old interface: */
static __inline__ __u16 U_OLD(const __u16 x)
{
	return x ^ 0xf000;
}
#ifndef __KERNEL__ /* only user side can have interest in the past data */
static __inline__ __u16 SUBMIT_UNI_OLD(const __u16 x)
{ 
	return U_OLD(x);
}
static __inline__ __u16 ACCEPT_UNI_OLD(const __u16 x)
{ 
	return U_OLD(x);
}
/* The problem with this was that public Unicodes U+f900--U+ffff used to be
 * mapped onto 0x0900--0x0f00 which collides with the area for (TYP,VAL) pairs
 */
#endif /* only user side */

#if 0 /* The old interface used to be equivalent to this: */

static __inline__ __u16 SUBMIT_UNI(const __u16 x)
{
	return SUBMIT_UNI_OLD(x);
}
static __inline__ __u16 ACCEPT_UNI(const __u16 x)
{
	return ACCEPT_UNI_OLD(x);
}
# ifdef __KERNEL__
#  define K_UNI_EXCHANGE_VERSION 0 /* will be explained below */
# endif

#else /* The new interface: */

#if (K_AREA_8BIT_BASE != 0xe000)
/* It depends on the assumption that (TYP,VAL) 8-bit pairs are stored 
 * in the 0xeNNN private subarea of Unicode. 
 * The dependency is due to the fact that SUBMIT_UNI will be used in vt.c
 * to return stored keymap values. Such a use is supposed to be possible 
 * in other applications too. */
#error Incompatible K_AREA_8BIT_BASE and SUBMIT_UNI/ACCEPT_UNI
#endif

#if 0

/* The possibly simplest correct way to define the conversion such that it
 * satisfies the requirement for (TYP,VAL) pairs "transparency" would be: */

/* 0xeNNN are private  -- use them as the area to hold the values */
static __inline__ __u16 U_CONVERSION(const __u16 x)
{
	return x ^ 0xe000; 
}
static __inline__ __u16 SUBMIT_UNI(const __u16 x)
{
	return U_CONVERSION(x);
}
static __inline__ __u16 ACCEPT_UNI(const __u16 x)
{ 
	return U_CONVERSION(x);
}
/* Its disadvantage is that it completely breaks the compatibility with the 
 * old interface on a mass of public Unicode values. */
# ifdef __KERNEL__
#  define K_UNI_EXCHANGE_VERSION 2 /* special number; anyway, not used so far */
# endif

#else 

/* This is another correct way, a more cumbersome one in the aspect of implementation. 
 * But its *advantage* is that it is very close to the old interface
 * even on most Unicode values: */
static __inline__ __u16 SUBMIT_UNI(const __u16 x)
{
	if ( (x & 0xe000) == 0xe000 )
		return U_OLD(x ^ 0x1000); /* swap 0xeNNN and 0xfNNN */
	else
		return U_OLD(x);
}
static __inline__ __u16 ACCEPT_UNI(const __u16 x)
{
	const __u16 pre = U_OLD(x);
	if ( (pre & 0xe000) == 0xe000 )
		return pre ^ 0x1000;  /* swap 0xeNNN and 0xfNNN back */
	else
		return pre;
}

/* Discussion.
 * So, if the user (loadkeys) uses the old interface conversion and
 * the kernel uses the new one (this one), then:
 * - user sends a (TYP,VAL) pair transparently, 
 *   kernel interprets it correctly;
 * - user sends a public Unicode < 0xe000, kernel interprets it correctly;
 * - public Unicode >= 0xf900, kernel interprets it as 0xe9NN (TYP,VAL), 
 *   FALSE! (but the old kernel also wasn't able to accept it, actually
 *   the behavior turns out to be euivavlent for this case for old and new);
 * No other codes could be sent by the old loadkeys.
 * There is almost no incompatibility for this combination.
 *
 * If the user uses the new interface, but the kernel is old, then:
 * - user sends a (TYP,VAL) pair transparently, 
 *   kernel interprets it correctly;
 * - user sends a public Unicode < 0xe000, kernel interprets it correctly;
 * - public Unicode >= 0xf900, kernel interprets it as 0xe9NN (private 
 *   unassigned) FALSE! (but not so bad; besides that, the old kernel didn't 
 *   have a way to accept these Unicodes);
 * - privately assigned Unicode from 0xf000..0xf8ff, which denotes 
 *   "CapsLock-able Letter with case", 
 *   kernel interprets it as 0xeNNN (private unassigned) FALSE! 
 *   (but not quite bad, since the old kernel didn't 
 *   have a way to store this kind of marked Unicodes);
 * The only really bad thing for this combination is the last 
 * incompatibility: an old loadkeys would simply ignore the 
 * "CapsLock-able Letter with case" mark for Unicodes, so they worked
 * as simple Unicode characters, but the combination of the new loadkeys
 * and an old kernel would cause them to be completely abandoned.
 * Probably, the new loadkeys program should check the version of the kernel
 * interface conversion before sending Unicodes with 
 * "CapsLock-able Letter with case" mark; the question is `How?'.
 *
 * My idea is to use a special reaction of the kernel
 * to a Unicode noncharcters ("These codes are intended for process internal 
 * uses, but are not permitted for interchange.")
 * as a means of indicating the exchange interface used by the kernel 
 * as follows. Range U+fdd0..U+fdef is noncharacters. If a user with new 
 * interface conversion sends U+fddN, the old kernel stores it as U+eddN
 * whereas the new kernel recieves it as U+fddN. The new kernel can either 
 * accept it or say EINVAL. Let it do that according to the following rules:
 * - let a byte (K_UNI_EXCHANGE_VERSION) be the number 
 *   of the interface version consisting of bits b7 b6 .. b0;
 * - if the user tries to submit U+fddN (0 <= N < 8), then
 *   the kernel says EINVAL iff bN == 1.
 * So, if our interface version is 1, then the kernel responds with EINVAL
 * to an attempt to map a key to U+fdd0.
 * The old kernel used to accept it.
 * It's up to loadkeys to decide on which key to test the version.
 * Keycode 0 is the most sage way since (acording to vt.c):
 * "assignment to entry 0 only tests validity of args" (for PC).
 *
 * Now let's turn our attention to the other side of the kernel-user program
 * communication: dumpkeys.
 * An old dumpkeys would surely write crap for values marked as
 * "CapsLock-able Letter with case"; well, we can do little about this.
 * But the combination of an old dumpkeys and the new kernel is fairly good:
 * most values will be understood correctly; as to the 
 * "CapsLock-able Letters with case", there are simply no such things in a 
 * an old kernel, so it doesn't bother us.
 */
# ifdef __KERNEL__
#  define K_UNI_EXCHANGE_VERSION 1 /* the version constant to report in vt.c */
# endif

#endif /* new variants */

#endif /* old/new variants */

#define K_AREA_TELL_EXCHVER_MASK 0xfff8
#define K_AREA_TELL_EXCHVER_BASE 0xfdd0
#define K_UNI_EXCH_VERSION_NR_BITS (0xffff & ~K_AREA_TELL_EXCHVER_MASK) /* must be a not more than 2-byte number */
#endif /* #ifndef _KBD_VALUES_H */
